/******************************************************************************
* Copyright 1996-2014 United States Government as represented by the
* Administrator of the National Aeronautics and Space Administration.
* All Rights Reserved.
******************************************************************************/
/******************************************************************************
*
*  NSSDC/CDF			CDF version/release/increment/sub-increment.
*
*  Version 1.0, 15-Dec-97, Hughes STX.
*
*  Modification history:
*
*   V1.0  15-Dec-97, J Love	Original version.  See the `CHANGES' file
*				for the changes which apply to a particular
*				version/release/increment/subincrement.
*   V2.7  22-Jul-99, M Liu	Change Version to V2.7.0
*         14-Apr-01, M Liu      Change Version to V2.7.1
*         08-Apr-04, M Liu      Change Version to V2.7.2
*   V3.0  22-Jun-04, M Liu      Change Version to V3.0.0
*   V3.1  19-May-05, M Liu      Change Version to V3.1.0
*   V3.2  10-May-06, M Liu      Change Version to V3.2.0
*   V3.2  24-Apr-08, M Liu      Change Version to V3.2.1
*         11-Aug-08, M Liu      Change Version to V3.2.2
*         20-Jan-09, M Liu      Change Version to V3.2.4
*   V3.3  21-Jan-09, M Liu      Change Version to V3.3.0
*         21-Jul-09, M Liu      Change Version to V3.3.1
*         25-Feb-11, M Liu      Change Version to V3.3.2
*   V3.4  16-Feb-12, M Liu      Change Version to V3.4.0
*   V3.5  02-Apr-12, M Liu      Change Version to V3.5.0
*   V3.6  02-Feb-15, M Liu      Change Version to V3.6.0
*         20-Sep-15, M Liu      Change Version to V3.6.1
******************************************************************************/

#if !defined(CDFREVh_INCLUDEd__)
#define CDFREVh_INCLUDEd__

#define CDF_LIBRARY_VERSION		(long) 3
#define CDF_LIBRARY_RELEASE		(long) 6
#define CDF_LIBRARY_INCREMENT		(long) 1
#define CDF_LIBRARY_subINCREMENT	'0'

#endif
